import socket
import time
from typing import *

import attr
import serial
import structlog

from elgas import exceptions

LOG = structlog.get_logger("transport")


class Timeout:
    """"""

    def __init__(self, duration):
        """Initialize a timeout with given duration"""
        self.is_infinite = duration is None
        self.is_non_blocking = duration == 0
        self.duration = duration
        if duration is not None:
            self.target_time = time.monotonic() + duration
        else:
            self.target_time = None

    def expired(self):
        """Return a boolean, telling if the timeout has expired"""
        return self.target_time is not None and self.time_left() <= 0

    def time_left(self):
        """Return how many seconds are left until the timeout expires"""
        if self.is_non_blocking:
            return 0
        elif self.is_infinite:
            return None
        else:
            delta = self.target_time - time.monotonic()
            if delta > self.duration:
                # clock jumped, recalculate
                self.target_time = time.monotonic() + self.duration
                return self.duration
            else:
                return max(0, delta)

    def restart(self, duration):
        """\
        Restart a timeout, only supported if a timeout was already set up
        before.
        """
        self.duration = duration
        self.target_time = time.monotonic() + duration


class ElgasTransport(Protocol):
    def connect(self):
        ...

    def disconnect(self):
        ...

    def send(self, data: bytes):
        ...

    def recv(self):
        ...


@attr.s(auto_attribs=True)
class BlockingTcpTransport:
    """
    A TCP transport using Blocking I/O.
    """

    host: str
    port: int
    timeout: int = attr.ib(default=10)
    tcp_socket: Optional[socket.socket] = attr.ib(init=False, default=None)

    @property
    def address(self) -> Tuple[str, int]:
        return self.host, self.port

    def connect(self):
        """
        Create a new socket and set it on the transport
        """
        if self.tcp_socket:
            raise RuntimeError(f"There is already an active socket to {self.address}")

        try:
            self.tcp_socket = socket.create_connection(
                address=self.address, timeout=self.timeout
            )
        except (
            OSError,
            IOError,
            socket.timeout,
            socket.error,
            ConnectionRefusedError,
        ) as e:
            raise exceptions.CommunicationError("Unable to connect socket") from e
        LOG.info(f"Connected to {self.address}")

    def disconnect(self):
        """
        Close socket and remove it from the transport. No-op if the socket is already
        closed.
        """
        if self.tcp_socket:
            # only disconnect if there is a socket.
            try:
                self.tcp_socket.shutdown(socket.SHUT_RDWR)
                self.tcp_socket.close()
            except (OSError, IOError, socket.timeout, socket.error) as e:
                self.tcp_socket = None
                raise exceptions.CommunicationError from e
            self.tcp_socket = None
            LOG.info(f"Connection to {self.address} is closed")

    def send(self, data: bytes):
        """"""
        if not self.tcp_socket:
            raise RuntimeError("TCP transport not connected.")
        try:
            self.tcp_socket.sendall(data)
            LOG.debug(f"Sent data", data=data, transport=self)
        except (OSError, IOError, socket.timeout, socket.error) as e:
            raise exceptions.CommunicationError("Could no send data") from e

    def recv(self) -> bytes:
        """"""
        try:
            data = self.recv_until()
            LOG.debug("Received data", data=data, transport=self)
        except (OSError, IOError, socket.timeout, socket.error) as e:
            raise exceptions.CommunicationError("Could not receive data") from e
        return data

    def _recv_bytes(self, amount: int):
        """
        Some implementations will return partial data, and we need to keep on trying
        to read the bytes until we have them all.
        """
        if not self.tcp_socket:
            raise RuntimeError("TCP transport not connected.")

        data = b""
        while len(data) < amount:
            data += self.tcp_socket.recv(amount - len(data))

        return data

    def recv_until(self, expected=b"\x0d", size=None):
        """\
        Read until an expected sequence is found ('\r' by default), the size
        is exceeded or until timeout occurs.
        """
        lenterm = len(expected)
        line = bytearray()
        timeout = Timeout(self.timeout)
        while True:
            c = self._recv_bytes(1)
            if c:
                line += c
                if line[-lenterm:] == expected:
                    break
                if size is not None and len(line) >= size:
                    break
            else:
                break
            if timeout.expired():
                break
        return bytes(line)


@attr.s(auto_attribs=True)
class SerialTransport:
    """
    A Serial transport
    """

    port: str
    baud_rate: int
    timeout: int = attr.ib(default=10)
    serial_port: Optional[serial.Serial] = attr.ib(init=False, default=None)

    def connect(self):
        """
        Create a new socket and set it on the transport
        """
        if self.serial_port:
            raise RuntimeError(f"There is already an active serial port")

        try:
            self.serial_port = serial.Serial(
                port=self.port,
                baudrate=self.baud_rate,
                timeout=self.timeout,
            )
        except (
            OSError,
            IOError,
            serial.SerialException,
        ) as e:
            raise exceptions.CommunicationError("Unable to connect socket") from e
        LOG.info(f"Connected to {self.port}")

    def disconnect(self):
        """
        Close socket and remove it from the transport. No-op if the socket is already
        closed.
        """
        if self.serial_port:
            # only disconnect if there is a socket.
            try:
                self.serial_port.close()
            except (OSError, IOError, serial.SerialException) as e:
                self.serial_port = None
                raise exceptions.CommunicationError from e
            self.serial_port = None
            LOG.info(f"Connection to {self.port} is closed")

    def send(self, data: bytes):
        """"""
        if not self.serial_port:
            raise RuntimeError("Serial transport not connected.")
        try:
            self.serial_port.write(data)
            LOG.debug(f"Sent data", data=data, transport=self)
        except (OSError, IOError, serial.SerialException) as e:
            raise exceptions.CommunicationError("Could no send data") from e

    def recv(self) -> bytes:
        """"""
        try:
            data = self.recv_until()
            LOG.debug("Received data", data=data, transport=self)
        except (OSError, IOError, serial.SerialException) as e:
            raise exceptions.CommunicationError("Could not receive data") from e
        return data

    def _recv_bytes(self, amount: int):
        """
        Some implementations will return partial data, and we need to keep on trying
        to read the bytes until we have them all.
        """
        if not self.serial_port:
            raise RuntimeError("TCP transport not connected.")

        data = b""
        while len(data) < amount:
            data += self.serial_port.read(amount - len(data))

        return data

    def recv_until(self, expected=b"\x0d", size=None):
        """\
        Read until an expected sequence is found ('\r' by default), the size
        is exceeded or until timeout occurs.
        """
        lenterm = len(expected)
        line = bytearray()
        timeout = Timeout(self.timeout)
        while True:
            c = self._recv_bytes(1)
            if c:
                line += c
                if line[-lenterm:] == expected:
                    break
                if size is not None and len(line) >= size:
                    break
            else:
                break
            if timeout.expired():
                break
        return bytes(line)



from __future__ import annotations

from typing import *

import attr

from elgas import constants, utils


@attr.s(auto_attribs=True)
class Request:
    TYPE: ClassVar[int] = 0x84
    service: constants.ServiceNumber
    destination_address_1: int
    destination_address_2: int
    source_address_1: int
    source_address_2: int
    data: bytes

    ...

    @property
    def length(self) -> int:
        """
        All bytes are counted except STX, before replacing reserved characters
        """
        return len(self.data) + 15

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(0x02)
        out.append(0xFE)
        out.append(self.TYPE)
        out.append(self.service)
        out.extend(self.length.to_bytes(2, "little"))
        out.extend(self.destination_address_1.to_bytes(2, "big"))
        out.append(self.destination_address_2)
        out.extend(self.source_address_1.to_bytes(2, "big"))
        out.append(self.source_address_2)
        out.extend(self.data)
        redundancy_source = out[1:]
        out.append(utils.calculate_lrc(redundancy_source))
        out.append(utils.calculate_checksum(redundancy_source))
        out.append(utils.calculate_drc(redundancy_source))
        out.append(0x0D)

        return bytes(out)


@attr.s(auto_attribs=True)
class Response:
    TYPE: ClassVar[int] = 0x86
    service: constants.ServiceNumber
    destination_address_1: int
    destination_address_2: int
    source_address_1: int
    source_address_2: int
    data: bytes

    @property
    def length(self) -> int:
        """
        All bytes are counted except STX, before replacing reserved characters
        """
        return len(self.data) + 15

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(0x02)
        out.append(0xFE)
        out.append(self.TYPE)
        out.append(self.service)
        out.extend(self.length.to_bytes(2, "little"))
        out.extend(self.destination_address_1.to_bytes(2, "little"))
        out.append(self.destination_address_2)
        out.extend(self.source_address_1.to_bytes(2, "little"))
        out.append(self.source_address_2)
        out.extend(self.data)
        redundancy_source = out[1:]
        out.append(utils.calculate_lrc(redundancy_source))
        out.append(utils.calculate_checksum(redundancy_source))
        out.append(utils.calculate_drc(redundancy_source))
        out.append(0x0D)

        return bytes(out)

    @classmethod
    def from_bytes(cls, in_bytes: bytes):
        if not in_bytes.startswith(b"\x02\xfe" + cls.TYPE.to_bytes(1, "little")):
            raise ValueError("Not a Response frame")
        if not in_bytes.endswith(b"\x0d"):
            raise ValueError("Data does not end with end char")
        service = constants.ServiceNumber(int.from_bytes(in_bytes[3:4], "little"))
        length = int.from_bytes(in_bytes[4:6], "little")
        if length != len(in_bytes) - 1:  # STX is not counted
            raise ValueError(
                f"Length field does not correspond to length of data. Got {len(in_bytes)-1}, should be {length} "
            )
        destination_address_1 = int.from_bytes(in_bytes[6:8], "little")
        destination_address_2 = int.from_bytes(in_bytes[8:9], "little")
        source_address_1 = int.from_bytes(in_bytes[9:11], "little")
        source_address_2 = int.from_bytes(in_bytes[11:12], "little")
        data = in_bytes[12:-4]
        redundancy_source = bytearray(in_bytes[1:-4])
        lrc = in_bytes[-4]
        calculated_lrc = utils.calculate_lrc(redundancy_source)
        if lrc != calculated_lrc:
            raise ValueError(
                f"Incorrect LRC. Got {lrc!r}, should be {calculated_lrc!r}"
            )
        checksum = in_bytes[-3]
        calculated_checksum = utils.calculate_checksum(redundancy_source)
        if checksum != calculated_checksum:
            raise ValueError(
                f"Incorrect CHECKSUM. Got {checksum!r}, should be {calculated_checksum!r}"
            )
        drc = in_bytes[-2]
        calculated_drc = utils.calculate_drc(redundancy_source)
        if drc != calculated_drc:
            raise ValueError(f"Incorrect DRC. Got {drc!r}, should be {drc!r}")

        return cls(
            service=service,
            destination_address_1=destination_address_1,
            destination_address_2=destination_address_2,
            source_address_1=source_address_1,
            source_address_2=source_address_2,
            data=data,
        )


@attr.s(auto_attribs=True)
class EncryptedRequest(Request):
    TYPE: ClassVar[int] = 0x85


@attr.s(auto_attribs=True)
class EncryptedResponse(Response):
    TYPE: ClassVar[int] = 0x87


@attr.s(auto_attribs=True)
class ResponseFactory:
    @staticmethod
    def from_bytes(in_data: bytes):
        if in_data.startswith(b"\x02\xfe\x86"):
            return Response.from_bytes(in_data)
        elif in_data.startswith(b"\x02\xfe\x87"):
            return EncryptedResponse.from_bytes(in_data)
        else:
            raise ValueError("Data is not a response frame")



import struct
from typing import ClassVar, Optional

import attr
import marshmallow

from elgas.parameters.enumerations import ParameterObjectType
from elgas.utils import pop_many, pretty_text


@attr.s(auto_attribs=True)
class SetPoint:
    object_type: ClassVar[ParameterObjectType] = ParameterObjectType.SET_POINT

    number: int
    id: int
    bit_order_in_actual_values: int
    bit_order_in_data_archive_record: int
    bit_order_in_binary_archive_record: int

    bit_control: int
    in_binary_archive: bool
    in_data_archive: bool
    active_indicator: bool
    name: str
    value_of_limit: float
    type_of_primary_quantity: int
    number_of_primary_quantity: int
    action_during_change: Optional[int]
    text_log_0: Optional[str]
    text_log_1: Optional[str]

    @classmethod
    def from_bytes(cls, in_data: bytes):
        data = bytearray(in_data)
        number = int.from_bytes(pop_many(data, 2), "little")
        id = int.from_bytes(pop_many(data, 2), "little")
        bit_order_in_actual_values = int.from_bytes(pop_many(data, 2), "little")
        bit_order_in_data_archive_record = int.from_bytes(pop_many(data, 2), "little")
        bit_order_in_binary_archive_record = int.from_bytes(pop_many(data, 2), "little")
        bit_control = data.pop(0)
        in_binary_archive = bool(bit_control & 0b00000001)
        in_data_archive = bool(bit_control & 0b00000010)
        active_indicator = bool(bit_control & 0b01000000)
        name = pretty_text(pop_many(data, 23))
        value_of_limit = struct.unpack("<f", pop_many(data, 4))[0]
        type_of_primary_quantity = data.pop(0)
        number_of_primary_quantity = data.pop(0)

        if data:
            action_during_change = data.pop(0)
            text_log_0 = pretty_text(pop_many(data, 13))
            text_log_1 = pretty_text(pop_many(data, 13))
        else:
            action_during_change = None
            text_log_0 = None
            text_log_1 = None

        assert not data

        return cls(
            number=number,
            id=id,
            bit_order_in_actual_values=bit_order_in_actual_values,
            bit_order_in_data_archive_record=bit_order_in_data_archive_record,
            bit_order_in_binary_archive_record=bit_order_in_binary_archive_record,
            bit_control=bit_control,
            in_binary_archive=in_binary_archive,
            in_data_archive=in_data_archive,
            active_indicator=active_indicator,
            name=name,
            value_of_limit=value_of_limit,
            type_of_primary_quantity=type_of_primary_quantity,
            number_of_primary_quantity=number_of_primary_quantity,
            action_during_change=action_during_change,
            text_log_0=text_log_0,
            text_log_1=text_log_1,
        )


class SetPointSchema(marshmallow.Schema):

    number = marshmallow.fields.Integer(required=True)
    id = marshmallow.fields.Integer(required=True)
    bit_order_in_actual_values = marshmallow.fields.Integer(required=True)
    bit_order_in_data_archive_record = marshmallow.fields.Integer(required=True)
    bit_order_in_binary_archive_record = marshmallow.fields.Integer(required=True)

    bit_control = marshmallow.fields.Integer(required=True)
    in_binary_archive = marshmallow.fields.Boolean(required=True)
    in_data_archive = marshmallow.fields.Boolean(required=True)
    active_indicator = marshmallow.fields.Boolean(required=True)
    name = marshmallow.fields.String(required=True)
    value_of_limit = marshmallow.fields.Float(required=True, as_string=True)
    type_of_primary_quantity = marshmallow.fields.Integer(required=True)
    number_of_primary_quantity = marshmallow.fields.Integer(required=True)
    action_during_change = marshmallow.fields.Integer(required=True, allow_none=True)
    text_log_0 = marshmallow.fields.String(required=True, allow_none=True)
    text_log_1 = marshmallow.fields.String(required=True, allow_none=True)



import struct
from typing import ClassVar, Optional

import attr
import marshmallow
from marshmallow import post_load

from elgas.parameters.enumerations import ParameterObjectType
from elgas.utils import pop_many, pretty_text


@attr.s(auto_attribs=True)
class AnalogQuantity:
    object_type: ClassVar[ParameterObjectType] = ParameterObjectType.ANALOG_MEASURAND
    value_length: ClassVar[int] = 2

    number: int
    id: int
    address_in_actual_values: int
    address_in_data_archive_record: int
    bit_control: int  # seems to be which archives the values is in?
    in_data_archive: bool
    in_daily_archive: bool
    in_monthly_archive: bool
    is_metrological_quantity: bool
    in_fast_archive_1: bool
    in_fast_archvie_2: bool
    name: str
    unit: str
    digit: float
    offset: float
    lower_limit_measuring_range: float
    upper_limit_measuring_range: float
    serial_number_transducer: int
    error_bit_order_in_actual_values: int
    error_bit_order_in_binary_archive: int
    error_bit_order_in_data_archive: int
    address_in_daily_archive_record: int
    address_in_monthly_archive_record: int
    samples_in_fast_archive: int
    decimals: Optional[int]

    @classmethod
    def from_bytes(cls, in_data: bytes):
        data = bytearray(in_data)
        number = int.from_bytes(pop_many(data, 2), "little")
        id = int.from_bytes(pop_many(data, 2), "little")
        address_in_actual_values = int.from_bytes(pop_many(data, 2), "little")
        address_in_data_archive_record = int.from_bytes(pop_many(data, 2), "little")
        bit_control = data.pop(0)
        in_data_archive = bool(bit_control & 0b00000001)
        in_daily_archive = bool(bit_control & 0b00000010)
        in_monthly_archive = bool(bit_control & 0b00000100)
        is_metrological_quantity = bool(bit_control & 0b00010000)
        in_fast_archive_1 = bool(bit_control & 0b00100000)
        in_fast_archvie_2 = bool(bit_control & 0b01000000)
        name = pretty_text(pop_many(data, 23))
        unit = pretty_text(pop_many(data, 8))
        digit = round(struct.unpack("<f", pop_many(data, 4))[0], 7)
        offset = struct.unpack("<f", pop_many(data, 4))[0]
        lower_limit_measuring_range = round(
            struct.unpack("<f", pop_many(data, 4))[0], 7
        )
        upper_limit_measuring_range = struct.unpack("<f", pop_many(data, 4))[0]
        serial_number_transducer = int.from_bytes(pop_many(data, 4), "little")
        error_bit_order_in_actual_values = int.from_bytes(pop_many(data, 2), "little")
        error_bit_order_in_binary_archive = int.from_bytes(pop_many(data, 2), "little")
        error_bit_order_in_data_archive = int.from_bytes(pop_many(data, 2), "little")
        address_in_daily_archive_record = int.from_bytes(pop_many(data, 2), "little")
        address_in_monthly_archive_record = int.from_bytes(pop_many(data, 2), "little")
        samples_in_fast_archive = data.pop(0)

        if data:
            decimals = data.pop(0)
        else:
            decimals = None

        return cls(
            number=number,
            id=id,
            address_in_actual_values=address_in_actual_values,
            address_in_data_archive_record=address_in_data_archive_record,
            bit_control=bit_control,
            in_data_archive=in_data_archive,
            in_daily_archive=in_daily_archive,
            in_monthly_archive=in_monthly_archive,
            is_metrological_quantity=is_metrological_quantity,
            in_fast_archive_1=in_fast_archive_1,
            in_fast_archvie_2=in_fast_archvie_2,
            name=name,
            unit=unit,
            digit=digit,
            offset=offset,
            lower_limit_measuring_range=lower_limit_measuring_range,
            upper_limit_measuring_range=upper_limit_measuring_range,
            serial_number_transducer=serial_number_transducer,
            error_bit_order_in_actual_values=error_bit_order_in_actual_values,
            error_bit_order_in_binary_archive=error_bit_order_in_binary_archive,
            error_bit_order_in_data_archive=error_bit_order_in_data_archive,
            address_in_daily_archive_record=address_in_daily_archive_record,
            address_in_monthly_archive_record=address_in_monthly_archive_record,
            samples_in_fast_archive=samples_in_fast_archive,
            decimals=decimals,
        )


class AnalogQuantitySchema(marshmallow.Schema):
    number = marshmallow.fields.Integer(required=True)
    id = marshmallow.fields.Integer(required=True)
    address_in_actual_values = marshmallow.fields.Integer(required=True)
    address_in_data_archive_record = marshmallow.fields.Integer(required=True)
    bit_control = marshmallow.fields.Integer(required=True)
    in_data_archive = marshmallow.fields.Boolean(required=True)
    in_daily_archive = marshmallow.fields.Boolean(required=True)
    in_monthly_archive = marshmallow.fields.Boolean(required=True)
    is_metrological_quantity = marshmallow.fields.Boolean(required=True)
    in_fast_archive_1 = marshmallow.fields.Boolean(required=True)
    in_fast_archvie_2 = marshmallow.fields.Boolean(required=True)
    name = marshmallow.fields.String(required=True)
    unit = marshmallow.fields.String(required=True)
    digit = marshmallow.fields.Number(required=True)
    offset = marshmallow.fields.Number(required=True)
    lower_limit_measuring_range = marshmallow.fields.Number(required=True)
    upper_limit_measuring_range = marshmallow.fields.Number(required=True)
    serial_number_transducer = marshmallow.fields.Integer(required=True)
    error_bit_order_in_actual_values = marshmallow.fields.Integer(required=True)
    error_bit_order_in_binary_archive = marshmallow.fields.Integer(required=True)
    error_bit_order_in_data_archive = marshmallow.fields.Integer(required=True)
    address_in_daily_archive_record = marshmallow.fields.Integer(required=True)
    address_in_monthly_archive_record = marshmallow.fields.Integer(required=True)
    samples_in_fast_archive = marshmallow.fields.Integer(required=True)
    decimals = marshmallow.fields.Integer(required=True, allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return AnalogQuantity(**data)



import datetime
from enum import IntEnum
from typing import *

import attr

from elgas import constants, utils


class Archive(IntEnum):
    EXTREME = 0
    DATA = 1
    BINARY = 2
    DAILY = 3
    MONTHLY = 4
    SETTING = 5
    FAST_1 = 6
    FAST_2 = 7
    STATUS = 8
    BILLING = 9
    GAS_COMPOSITION = 10


@attr.s(auto_attribs=True)
class ReadInstantaneousValuesRequest:
    """
    Will request a readout of all current parameters in the device
    """

    service: ClassVar[constants.ServiceNumber] = constants.ServiceNumber.READ_VALUES
    password: str

    def to_bytes(self) -> bytes:
        return utils.pad_password(self.password).encode("latin-1")


@attr.s(auto_attribs=True)
class ReadInstantaneousValuesResponse:
    """
    A long datastructure of parameters that depends on the firmware of the device.
    The parsing is set to ELCORplus Gen 4.
    Instantaneous values are called Actual Values in documentation.
    """

    service: ClassVar[constants.ServiceNumber] = constants.ServiceNumber.READ_VALUES
    current_time: datetime
    is_dst: bool
    supports_dst: bool
    data: bytes
    data_access: int
    status: int
    summary_status: int
    parameter_crc: bytes

    @classmethod
    def from_bytes(cls, in_bytes: bytes):
        data = in_bytes
        current_time, is_dst, supports_dst = utils.bytes_to_datetime(data[:6])
        data = data[6:]
        parameter_crc = data[-2:]
        data = data[:-2]
        summary_status = data[-8:]
        data = data[:-8]
        status = data[-8:]
        data = data[:-8]
        data_access = data[-1]
        data = data[:-1]
        return cls(
            data=data,
            current_time=current_time,
            is_dst=is_dst,
            supports_dst=supports_dst,
            parameter_crc=parameter_crc,
            summary_status=summary_status,
            status=status,
            data_access=data_access,
        )


@attr.s(auto_attribs=True)
class ReadTimeRequest:
    """
    Request for the device's time. It contains no data.
    """

    service: ClassVar[
        constants.ServiceNumber
    ] = constants.ServiceNumber.READ_DEVICE_TIME

    def to_bytes(self) -> bytes:
        return b""


@attr.s(auto_attribs=True)
class ReadTimeResponse:
    """
    Contains the device time. Some firmware adds internal data at the end of the
    message.
    """

    service: ClassVar[
        constants.ServiceNumber
    ] = constants.ServiceNumber.READ_DEVICE_TIME

    time: datetime
    data_access_result: bytes
    is_dst: bool
    device_supports_dst: bool
    extra_data: bytes

    @classmethod
    def from_bytes(cls, in_bytes: bytes):
        time_data = in_bytes[:6]
        data_access_result = in_bytes[6:7]
        extra_data = in_bytes[7:-2]
        # crc = in_bytes[-2:]
        # TODO: Check CRC
        device_time, is_dst, supports_dst = utils.bytes_to_datetime(time_data)
        return cls(
            time=device_time,
            data_access_result=data_access_result,
            is_dst=is_dst,
            device_supports_dst=supports_dst,
            extra_data=extra_data,
        )


@attr.s(auto_attribs=True)
class WriteTimeRequest:
    """
    Request for the device's time. It contains no data.
    """

    service: ClassVar[
        constants.ServiceNumber
    ] = constants.ServiceNumber.WRITE_DEVICE_TIME

    password: str
    device_time: datetime

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.extend(utils.pad_password(self.password).encode("latin-1"))
        out.extend(utils.datetime_to_bytes(self.device_time))
        out.append(0b00000100)  # Flag to only sync time.

        return bytes(out)


@attr.s(auto_attribs=True)
class WriteTimeResponse:
    """
    Contains the device time. Some firmware adds internal data at the end of the
    message.
    """

    service: ClassVar[
        constants.ServiceNumber
    ] = constants.ServiceNumber.READ_DEVICE_TIME

    @classmethod
    def from_bytes(cls, in_bytes: bytes):

        return cls()


@attr.s(auto_attribs=True)
class ReadDeviceParametersRequest:
    """"""

    service: ClassVar[
        constants.ServiceNumber
    ] = constants.ServiceNumber.READ_SCADA_PARAMETERS

    password: str  # todo: set max lenght
    object_count: int  # todo: set max and min
    buffer_length: int  # todo: set max and min

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.extend(utils.pad_password(self.password).encode("latin-1"))
        out.extend(self.object_count.to_bytes(2, "little"))
        out.extend(self.buffer_length.to_bytes(2, "little"))

        return bytes(out)


@attr.s(auto_attribs=True)
class ReadDeviceParametersResponse:
    """"""

    service: ClassVar[
        constants.ServiceNumber
    ] = constants.ServiceNumber.READ_SCADA_PARAMETERS

    data: bytes
    object_number: int
    object_amount: int
    is_end: int  # TODO: Should this be a separate class to use in state handling?

    @classmethod
    def from_bytes(cls, in_bytes: bytes):
        object_number = int.from_bytes(in_bytes[:2], "little")
        object_amount = int.from_bytes(in_bytes[2:4], "little")
        is_end = bool(in_bytes[4])
        data = in_bytes[5:]
        return cls(
            object_number=object_number,
            object_amount=object_amount,
            is_end=is_end,
            data=data,
        )


@attr.s(auto_attribs=True)
class ReadArchiveByTimeRequest:
    """"""

    service: ClassVar[
        constants.ServiceNumber
    ] = constants.ServiceNumber.READ_ARCHIVES_BY_DATE

    password: str  # todo: set max lenght
    archive: constants.Archive
    amount: int
    oldest_timestamp: datetime

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.extend(utils.pad_password(self.password).encode("latin-1"))
        out.append(self.archive)
        out.extend(self.amount.to_bytes(2, "little"))
        out.extend(utils.datetime_to_bytes(self.oldest_timestamp))

        return bytes(out)


@attr.s(auto_attribs=True)
class ReadArchiveByTimeResponse:
    """"""

    service: ClassVar[
        constants.ServiceNumber
    ] = constants.ServiceNumber.READ_ARCHIVES_BY_DATE

    archive: constants.Archive
    oldest_record_id: int
    data: bytes

    @classmethod
    def from_bytes(cls, in_bytes: bytes):
        archive = constants.Archive(in_bytes[0])
        oldest_record_id = int.from_bytes(in_bytes[1:4], "little")
        data = in_bytes[5:]
        return cls(
            archive=archive,
            oldest_record_id=oldest_record_id,
            data=data,
        )


@attr.s(auto_attribs=True)
class ReadArchiveRequest:
    """"""

    service: ClassVar[constants.ServiceNumber] = constants.ServiceNumber.READ_ARCHIVES

    password: str  # todo: set max lenght
    archive: constants.Archive
    amount: int
    oldest_record_id: int

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.extend(utils.pad_password(self.password).encode("latin-1"))
        out.append(self.archive)
        out.extend(self.oldest_record_id.to_bytes(4, "little"))
        out.extend(self.amount.to_bytes(2, "little"))

        return bytes(out)


@attr.s(auto_attribs=True)
class ReadArchiveResponse:
    """"""

    service: ClassVar[constants.ServiceNumber] = constants.ServiceNumber.READ_ARCHIVES

    archive: constants.Archive
    oldest_record_id: int
    data: bytes

    @classmethod
    def from_bytes(cls, in_bytes: bytes):
        archive = constants.Archive(in_bytes[0])
        oldest_record_id = int.from_bytes(in_bytes[1:4], "little")
        data = in_bytes[5:]
        return cls(
            archive=archive,
            oldest_record_id=oldest_record_id,
            data=data,
        )



from typing import ClassVar, Dict, Type

import attr

from elgas.parameters.analog_quantity import AnalogQuantity, AnalogQuantitySchema
from elgas.parameters.binary import Binary, BinarySchema
from elgas.parameters.compressibility import (
    Compressibility,
    CompressibilitySchema,
    CompressibilityZ,
    CompressibilityZBase,
    CompressibilityZBaseSchema,
    CompressibilityZSchema,
)
from elgas.parameters.conversion_coefficient import (
    ConversionCoefficient,
    ConversionCoefficientSchema,
)
from elgas.parameters.counter import (
    Counter,
    CounterSchema,
    DoubleCounter,
    DoubleCounterSchema,
)
from elgas.parameters.device_error import DeviceError, DeviceErrorSchema
from elgas.parameters.diagnostics import Diagnostics, DiagnosticsSchema
from elgas.parameters.difference_counter import (
    DifferenceBaseCounter,
    DifferenceBaseCounterSchema,
    DifferenceCounter,
    DifferenceCounterSchema,
)
from elgas.parameters.energy import Energy, EnergySchema, ErrorEnergy, ErrorEnergySchema
from elgas.parameters.enumerations import ParameterObjectType
from elgas.parameters.error_counter import (
    DoubleErrorCounter,
    DoubleErrorCounterSchema,
    ErrorCounter,
    ErrorCounterSchema,
)
from elgas.parameters.error_standard_counter import (
    ErrorStandardCounter,
    ErrorStandardCounterSchema,
)
from elgas.parameters.flow_rate import (
    FlowRate,
    FlowRateSchema,
    StandardFlowRate,
    StandardFlowRateSchema,
)
from elgas.parameters.modem import Modem, ModemSchema
from elgas.parameters.setpoint import SetPoint, SetPointSchema
from elgas.parameters.standard_counter import StandardCounter, StandardCounterSchema
from elgas.parameters.statistics import (
    AnalogStatistics,
    AnalogStatisticsSchema,
    AnalogTimeStatistics,
    AnalogTimeStatisticsSchema,
    CounterStatistics,
    CounterStatisticsSchema,
    StandardCounterStatistics,
    StandardCounterStatisticsSchema,
    Statistics,
    StatisticsSchema,
    TimeStatistics,
    TimeStatisticsSchema,
)
from elgas.parameters.sum_of_alarms import SumOfAlarms, SumOfAlarmsSchema
from elgas.parameters.system_parameters import SystemParameters, SystemParametersSchema
from elgas.parameters.tariff_counter import (
    BaseTariffCounter,
    BaseTariffCounterSchema,
    DoubleTariffCounter,
    DoubleTariffCounterSchema,
    TariffCounter,
    TariffCounterSchema,
)
from elgas.parameters.time_window import TimeWindow, TimeWindowSchema
from elgas.parameters.timer import Timer, TimerSchema


@attr.s(auto_attribs=True)
class ParameterFactory:

    object_map: ClassVar[Dict[ParameterObjectType, Type]] = {
        ParameterObjectType.SYSTEM_PARAMETER: SystemParameters,
        ParameterObjectType.ANALOG_MEASURAND: AnalogQuantity,
        ParameterObjectType.BINARY: Binary,
        ParameterObjectType.COUNTER: Counter,
        ParameterObjectType.STANDARD_COUNTER: StandardCounter,
        ParameterObjectType.FLOW_RATE: FlowRate,
        ParameterObjectType.STANDARD_FLOW_RATE: StandardFlowRate,
        ParameterObjectType.CONVERSION_COEFFICIENT: ConversionCoefficient,
        ParameterObjectType.ERROR_COUNTER: ErrorCounter,
        ParameterObjectType.ERROR_STANDARD_COUNTER: ErrorStandardCounter,
        ParameterObjectType.COMPRESSIBILITY: Compressibility,
        ParameterObjectType.TIME_WINDOW: TimeWindow,
        ParameterObjectType.DOUBLE_COUNTER: DoubleCounter,
        ParameterObjectType.DOUBLE_ERROR_COUNTER: DoubleErrorCounter,
        ParameterObjectType.DIAGNOSTICS: Diagnostics,
        ParameterObjectType.SET_POINT: SetPoint,
        ParameterObjectType.COMPRESSIBILITY_Z: CompressibilityZ,
        ParameterObjectType.COMPRESSIBILITY_Z_BASE: CompressibilityZBase,
        ParameterObjectType.ENERGY: Energy,
        ParameterObjectType.ERROR_ENERGY: ErrorEnergy,
        ParameterObjectType.TARIFF_COUNTER: TariffCounter,
        ParameterObjectType.DOUBLE_TARIFF_COUNTER: DoubleTariffCounter,
        ParameterObjectType.BASE_TARIFF_COUNTER: BaseTariffCounter,
        ParameterObjectType.DIFFERENCE_COUNTER: DifferenceCounter,
        ParameterObjectType.DIFFERENCE_BASE_COUNTER: DifferenceBaseCounter,
        ParameterObjectType.TIMER: Timer,
        ParameterObjectType.DEVICE_ERROR: DeviceError,
        ParameterObjectType.SUM_OF_ALARMS: SumOfAlarms,
        ParameterObjectType.ANALOG_STATISTICS: AnalogStatistics,
        ParameterObjectType.STATISTICS: Statistics,
        ParameterObjectType.ANALOG_TIME_STATISTICS: AnalogTimeStatistics,
        ParameterObjectType.TIME_STATISTICS: TimeStatistics,
        ParameterObjectType.COUNTER_STATISTICS: CounterStatistics,
        ParameterObjectType.STANDARD_COUNTER_STATISTICS: StandardCounterStatistics,
        ParameterObjectType.MODEM: Modem,
    }

    @staticmethod
    def from_bytes(object_type: ParameterObjectType, in_data: bytes):
        klass = ParameterFactory.object_map[object_type]
        return klass.from_bytes(in_data)


@attr.s(auto_attribs=True)
class ParameterSchemaFactory:

    object_map: ClassVar[Dict[ParameterObjectType, Type]] = {
        ParameterObjectType.SYSTEM_PARAMETER: SystemParametersSchema,
        ParameterObjectType.ANALOG_MEASURAND: AnalogQuantitySchema,
        ParameterObjectType.BINARY: BinarySchema,
        ParameterObjectType.COUNTER: CounterSchema,
        ParameterObjectType.STANDARD_COUNTER: StandardCounterSchema,
        ParameterObjectType.FLOW_RATE: FlowRateSchema,
        ParameterObjectType.STANDARD_FLOW_RATE: StandardFlowRateSchema,
        ParameterObjectType.CONVERSION_COEFFICIENT: ConversionCoefficientSchema,
        ParameterObjectType.ERROR_COUNTER: ErrorCounterSchema,
        ParameterObjectType.ERROR_STANDARD_COUNTER: ErrorStandardCounterSchema,
        ParameterObjectType.COMPRESSIBILITY: CompressibilitySchema,
        ParameterObjectType.TIME_WINDOW: TimeWindowSchema,
        ParameterObjectType.DOUBLE_COUNTER: DoubleCounterSchema,
        ParameterObjectType.DOUBLE_ERROR_COUNTER: DoubleErrorCounterSchema,
        ParameterObjectType.DIAGNOSTICS: DiagnosticsSchema,
        ParameterObjectType.SET_POINT: SetPointSchema,
        ParameterObjectType.COMPRESSIBILITY_Z: CompressibilityZSchema,
        ParameterObjectType.COMPRESSIBILITY_Z_BASE: CompressibilityZBaseSchema,
        ParameterObjectType.ENERGY: EnergySchema,
        ParameterObjectType.ERROR_ENERGY: ErrorEnergySchema,
        ParameterObjectType.TARIFF_COUNTER: TariffCounterSchema,
        ParameterObjectType.DOUBLE_TARIFF_COUNTER: DoubleTariffCounterSchema,
        ParameterObjectType.BASE_TARIFF_COUNTER: BaseTariffCounterSchema,
        ParameterObjectType.DIFFERENCE_COUNTER: DifferenceCounterSchema,
        ParameterObjectType.DIFFERENCE_BASE_COUNTER: DifferenceBaseCounterSchema,
        ParameterObjectType.TIMER: TimerSchema,
        ParameterObjectType.DEVICE_ERROR: DeviceErrorSchema,
        ParameterObjectType.SUM_OF_ALARMS: SumOfAlarmsSchema,
        ParameterObjectType.ANALOG_STATISTICS: AnalogStatisticsSchema,
        ParameterObjectType.STATISTICS: StatisticsSchema,
        ParameterObjectType.ANALOG_TIME_STATISTICS: AnalogTimeStatisticsSchema,
        ParameterObjectType.TIME_STATISTICS: TimeStatisticsSchema,
        ParameterObjectType.COUNTER_STATISTICS: CounterStatisticsSchema,
        ParameterObjectType.STANDARD_COUNTER_STATISTICS: StandardCounterStatisticsSchema,
        ParameterObjectType.MODEM: ModemSchema,
    }

    @staticmethod
    def by_parameter_id(parameter_id: ParameterObjectType):
        return ParameterSchemaFactory.object_map[parameter_id]



from __future__ import annotations

from typing import *

import attr
import structlog

from elgas import application, constants, exceptions, frames, security, state, utils

LOG = structlog.get_logger("connection")


@attr.s(auto_attribs=True)
class ResponsePduFactory:
    """
    Creates PDU classes from Elgas Responses via the service number.
    """

    service_map: ClassVar[Mapping] = {
        constants.ServiceNumber.READ_VALUES: application.ReadInstantaneousValuesResponse,
        constants.ServiceNumber.READ_DEVICE_TIME: application.ReadTimeResponse,
        constants.ServiceNumber.READ_SCADA_PARAMETERS: application.ReadDeviceParametersResponse,
        constants.ServiceNumber.READ_ARCHIVES_BY_DATE: application.ReadArchiveByTimeResponse,
        constants.ServiceNumber.READ_ARCHIVES: application.ReadArchiveResponse,
        constants.ServiceNumber.WRITE_DEVICE_TIME: application.WriteTimeResponse,
    }

    @staticmethod
    def from_response(response: frames.Response):
        klass = ResponsePduFactory.service_map[response.service]
        return klass.from_bytes(response.data)


def create_cipher_context(
    connection: "ElgasConnection",
) -> Optional[security.CipherContext]:
    """
    If both encryption key and encryption key id is set then return a cipher context
    """
    if connection.encryption_key is None or connection.encryption_key_id is None:
        return None
    else:
        return security.CipherContext(
            connection.encryption_key_id, connection.encryption_key
        )


def raise_error(data: bytes):
    error = data[0]
    if error == 0b00000001:
        raise exceptions.WrongPasswordError()
    elif error == 0b00000010:
        raise exceptions.SettingArchiveFull()
    elif error == 0b00000100:
        raise exceptions.SwitchOff()
    elif error == 0b00001000:
        raise exceptions.BlockedBuffer()
    elif error == 0b00010000:
        raise exceptions.DataError()
    elif error == 0b00100000:
        raise exceptions.CipherKeyError()
    elif error == 0b01000000:
        raise exceptions.WrongEncryptionKeysError()
    elif error == 0b10000000:
        raise exceptions.WriteError()


@attr.s(auto_attribs=True)
class ElgasConnection:
    """
    address1 is used for addressing measuring place  0-65535
    address2 is used for addressing concrete measuring place 0-255
    TODO: exactly what does that mean?..
    Address 0 in any will make all devices respond. In the response the correct
    address will be sent.


    """

    source_address_1: int = attr.ib(
        validator=[attr.validators.ge(0), attr.validators.le(65535)]
    )
    source_address_2: int = attr.ib(
        validator=[attr.validators.ge(0), attr.validators.le(255)]
    )
    destination_address_1: int = attr.ib(
        validator=[attr.validators.ge(0), attr.validators.le(65535)]
    )
    destination_address_2: int = attr.ib(
        validator=[attr.validators.ge(0), attr.validators.le(255)]
    )

    password: str = attr.ib(validator=[attr.validators.max_len(6)])
    password_id: int = attr.ib(
        validator=[attr.validators.ge(801), attr.validators.le(849)]
    )

    encryption_key: Optional[bytes] = attr.ib(default=None)
    encryption_key_id: Optional[security.EncryptionKeyId] = attr.ib(default=None)
    # TODO: Validate that there is an encryption key id if there is an encryption key.

    buffer: bytearray = attr.ib(init=False, factory=bytearray)
    elgas_state: state.ElgasState = attr.ib(factory=state.ElgasState)
    cipher_context: Optional[security.CipherContext] = attr.ib(
        default=attr.Factory(lambda self: create_cipher_context(self), takes_self=True)
    )

    def send(self, event) -> bytes:

        LOG.debug("Sending request PDU", pdu=event)
        self.elgas_state.process_event(event)
        application_data = event.to_bytes()

        if self.cipher_context:
            # replace application_data with encrypted data.

            request = frames.EncryptedRequest(
                service=event.service,
                destination_address_1=self.destination_address_1,
                destination_address_2=self.destination_address_2,
                source_address_1=self.source_address_1,
                source_address_2=self.source_address_2,
                data=self.cipher_context.encrypt(application_data),
            )

        else:
            request = frames.Request(
                service=event.service,
                destination_address_1=self.destination_address_1,
                destination_address_2=self.destination_address_2,
                source_address_1=self.source_address_1,
                source_address_2=self.source_address_2,
                data=application_data,
            )

        LOG.debug("Sending ELGAS request", request=request)
        frame_data = request.to_bytes()
        escaped_data = utils.escape_characters(frame_data)
        return escaped_data

    def receive_data(self, data: bytes):
        """
        Add data into the receive buffer. Also deescape the data so correct data is in
        the buffer.
        After this you could call next_event
        """
        if data:
            if data == b"\x0d" and len(self.buffer) == 0:
                LOG.info(
                    "Received an end char when read buffer was empty. This is a"
                    " known issue and data is ignored."
                )
                return

            self.buffer += data
            LOG.debug(
                f"Added data to connection buffer", data=data, total_buffer=self.buffer
            )

    def next_event(self) -> bytes:

        if not self.buffer:
            LOG.debug("Buffer empty. Need more data")
            return state.NEED_DATA

        # do we have an end char?
        try:
            end_char_index = self.buffer.index(b"\x0d") + 1
        except ValueError:
            # No end char in buffer and we need more data.
            LOG.debug("No end char in data. Need more data")
            return state.NEED_DATA

        data = self.buffer[:end_char_index]
        try:
            response = frames.ResponseFactory.from_bytes(utils.return_characters(data))

            LOG.debug("Received ELGAS response", response=response)

            if isinstance(response, frames.EncryptedResponse):
                response.data = self.cipher_context.decrypt(response.data)
                LOG.debug("Decrypted ELGAS response", decrypted=response)

            if len(response.data) == 1:
                raise_error(response.data)

        except ValueError:
            LOG.debug("Cannot interpret data as frame. Need more data")
            return state.NEED_DATA

        pdu = ResponsePduFactory.from_response(response)
        LOG.info("Received PDU", pdu=pdu)
        self.elgas_state.process_event(pdu)

        self.buffer = bytearray()  # clear buffer

        return pdu



from enum import IntEnum


class ServiceNumber(IntEnum):
    READ_VALUES = 0x64
    WRITE_VALUES = 0x65
    SEARCH_ARCHIVE_POINTERS_OLD = 0x67
    SEARCH_ARCHIVE_POINTERS = 0x7B
    READ_ARCHIVES_OLD = 0x68
    READ_ARCHIVES = 0x7C
    READ_DEVICE_TIME = 0x6C
    READ_SCADA_PARAMETERS = 0x70
    WRITE_DEVICE_TIME = 0x71
    GROUP_WRITE_VALUES = 0x77
    GROUP_READ_VALUES = 0x79
    CALL = 0x87
    READ_ARCHIVES_BY_DATE_OLD = 0x91
    READ_ARCHIVES_BY_DATE = 0x7D


class Archive(IntEnum):
    EXTREME = 0
    DATA = 1
    BINARY = 2
    DAILY = 3
    MONTHLY = 4
    SETTING = 5
    FAST_1 = 6
    FAST_2 = 7
    STATUS = 8
    BILLING = 9
    GAS_COMPOSITION = 10



import datetime
from typing import *

import attr
import structlog

from elgas import application, connection, constants, parser, state, transport, utils

LOG = structlog.get_logger("client")


@attr.s(auto_attribs=True)
class ElgasClient:
    """Elgas client"""

    transport: transport.ElgasTransport
    password: str = attr.ib(converter=utils.to_secret_str)
    password_id: int
    encryption_key: Optional[bytes] = attr.ib(
        converter=utils.to_secret_byte, default=None
    )
    encryption_key_id: Optional[int] = attr.ib(default=None)
    # device_configuration: List[object]
    elgas_connection: connection.ElgasConnection = attr.ib(
        default=attr.Factory(
            lambda self: connection.ElgasConnection(
                password=self.password,
                password_id=self.password_id,
                destination_address_1=0,
                destination_address_2=0,
                source_address_1=0,
                source_address_2=0,
                encryption_key=self.encryption_key,
                encryption_key_id=self.encryption_key_id,
            ),
            takes_self=True,
        )
    )

    def connect(self):
        self.transport.connect()

    def disconnect(self):
        self.transport.disconnect()

    def send(self, *events):
        for event in events:
            data = self.elgas_connection.send(event)
            self.transport.send(data)
            response_data = self.transport.recv()
            self.elgas_connection.receive_data(response_data)

    def next_event(self) -> Any:
        while True:
            event = self.elgas_connection.next_event()
            if event is state.NEED_DATA:
                self.elgas_connection.receive_data(self.transport.recv())
                continue
            return event

    def read_instantaneous_values(self) -> application.ReadInstantaneousValuesResponse:
        LOG.info("Reading instantaneous values")
        request = application.ReadInstantaneousValuesRequest(password=self.password)
        self.send(request)
        response: application.ReadInstantaneousValuesResponse = self.next_event()
        LOG.info("Received instantaneous values")
        return response

    def read_time(self):
        LOG.info("Reading device time")
        request = application.ReadTimeRequest()
        self.send(request)
        response: application.ReadTimeResponse = self.next_event()
        LOG.info("Got device time", time=response.time.isoformat())
        return response.time

    def write_time(self, device_time: datetime):
        LOG.info("Writing time to device", time=device_time.isoformat())
        request = application.WriteTimeRequest(
            password=self.password, device_time=device_time
        )
        self.send(request)
        self.next_event()
        LOG.info("Finished writing time to device")

    def read_parameters(self, read_from: int = 0) -> List[Any]:
        LOG.info(f"Reading device parameters", read_from=read_from)
        should_stop = False
        total_parameter_data = b""
        object_count = read_from
        while not should_stop:
            LOG.info("Requesting device parameters", object_count=object_count)
            request = application.ReadDeviceParametersRequest(
                password=self.password, object_count=object_count, buffer_length=1024
            )
            self.send(request)
            response: application.ReadDeviceParametersResponse = self.next_event()
            LOG.info("Received device parameters", object_amount=response.object_amount)
            total_parameter_data += response.data
            should_stop = response.is_end
            object_count += response.object_amount
            if not should_stop:
                LOG.info("More parameters available", next_object_count=object_count)

        LOG.debug("Received total parameters data", data=total_parameter_data)
        parsed = parser.ScadaParameterParser().parse(total_parameter_data)
        return parsed

    def read_archive(
        self, archive: constants.Archive, amount: int, oldest_record_id: int
    ):
        request = application.ReadArchiveRequest(
            password=self.password,
            archive=archive,
            amount=amount,
            oldest_record_id=oldest_record_id,
        )
        self.send(request)
        response = self.next_event()
        return response

    def read_archive_by_time(
        self, archive: constants.Archive, amount: int, oldest_timestamp: datetime
    ):
        request = application.ReadArchiveByTimeRequest(
            password=self.password,
            archive=archive,
            amount=amount,
            oldest_timestamp=oldest_timestamp,
        )
        self.send(request)
        response = self.next_event()
        return response



from typing import ClassVar, Optional

import attr
import marshmallow

from elgas.parameters.enumerations import ParameterObjectType
from elgas.utils import parse_ip_address, pop_many, pretty_text


@attr.s(auto_attribs=True)
class Modem:
    object_type: ClassVar[ParameterObjectType] = ParameterObjectType.MODEM

    number: int
    bit_control_0: int
    title: str
    modem_type: int
    initialization: str
    call_to_dispatching: str
    modem_hang_up: str
    special_initialization: str
    ip_address_for_registration_and_diagnostics: str
    ip_address_for_calling_to_dispatching: str
    registration_send_period: int
    authentication_mode: int
    port_for_registration: int
    port_for_calling_to_dispatching: int
    sms_call: str
    gprs_user_name: str
    gprs_password: str
    ip_address_for_ping: str
    ping_period: int
    transition_into_command_mode: str
    pin: str
    owner_sim_number: Optional[str]

    @classmethod
    def from_bytes(cls, in_data: bytes):
        data = bytearray(in_data)
        number = data.pop(0)
        bit_control_0 = data.pop(0)
        title = pretty_text(pop_many(data, 7))
        modem_type = data.pop(0)
        initialization = pretty_text(pop_many(data, 32))
        call_to_dispatching = pretty_text(pop_many(data, 32))
        modem_hang_up = pretty_text(pop_many(data, 8))
        special_initialization = pretty_text(pop_many(data, 80))
        ip_address_for_registration_and_diagnostics = parse_ip_address(
            pop_many(data, 4)
        )
        ip_address_for_calling_to_dispatching = parse_ip_address(pop_many(data, 4))
        registration_send_period = data.pop(0)
        authentication_mode = data.pop(0)
        port_for_registration = int.from_bytes(pop_many(data, 2), "little")
        port_for_calling_to_dispatching = int.from_bytes(pop_many(data, 2), "little")
        sms_call = pretty_text(pop_many(data, 32))
        gprs_user_name = pretty_text(pop_many(data, 49))
        gprs_password = pop_many(data, 33).hex()
        ip_address_for_ping = parse_ip_address(pop_many(data, 4))
        ping_period = int.from_bytes(pop_many(data, 2), "little")
        transition_into_command_mode = pretty_text(pop_many(data, 8))
        pin = pop_many(data, 9).hex()
        owner_sim_number = pretty_text(pop_many(data, 17))

        assert not data  # Should be empty now

        return cls(
            number=number,
            bit_control_0=bit_control_0,
            title=title,
            modem_type=modem_type,
            initialization=initialization,
            call_to_dispatching=call_to_dispatching,
            modem_hang_up=modem_hang_up,
            special_initialization=special_initialization,
            ip_address_for_registration_and_diagnostics=ip_address_for_registration_and_diagnostics,
            ip_address_for_calling_to_dispatching=ip_address_for_calling_to_dispatching,
            registration_send_period=registration_send_period,
            authentication_mode=authentication_mode,
            port_for_registration=port_for_registration,
            port_for_calling_to_dispatching=port_for_calling_to_dispatching,
            sms_call=sms_call,
            gprs_user_name=gprs_user_name,
            gprs_password=gprs_password,
            ip_address_for_ping=ip_address_for_ping,
            ping_period=ping_period,
            transition_into_command_mode=transition_into_command_mode,
            pin=pin,
            owner_sim_number=owner_sim_number,
        )


class ModemSchema(marshmallow.Schema):

    number = marshmallow.fields.Integer(required=True)
    bit_control_0 = marshmallow.fields.Integer(required=True)
    title = marshmallow.fields.String(required=True)
    modem_type = marshmallow.fields.Integer(required=True)
    initialization = marshmallow.fields.String(required=True)
    call_to_dispatching = marshmallow.fields.String(required=True)
    modem_hang_up = marshmallow.fields.String(required=True)
    special_initialization = marshmallow.fields.String(required=True)
    ip_address_for_registration_and_diagnostics = marshmallow.fields.String(
        required=True
    )
    ip_address_for_calling_to_dispatching = marshmallow.fields.String(required=True)
    registration_send_period = marshmallow.fields.Integer(required=True)
    authentication_mode = marshmallow.fields.Integer(required=True)
    port_for_registration = marshmallow.fields.Integer(required=True)
    port_for_calling_to_dispatching = marshmallow.fields.Integer(required=True)
    sms_call = marshmallow.fields.String(required=True)
    gprs_user_name = marshmallow.fields.String(required=True)
    gprs_password = marshmallow.fields.String(required=True)
    ip_address_for_ping = marshmallow.fields.String(required=True)
    ping_period = marshmallow.fields.Integer(required=True)
    transition_into_command_mode = marshmallow.fields.String(required=True)
    pin = marshmallow.fields.String(required=True)
    owner_sim_number = marshmallow.fields.String(required=True, allow_none=True)



from typing import *

import attr
import marshmallow


@attr.s(auto_attribs=True)
class ConfigurationObject:

    parameter_type: int
    data: Dict


class ConfigurationObjectSchema(marshmallow.Schema):

    parameter_type = marshmallow.fields.Integer(required=True)
    data = marshmallow.fields.Dict(required=True)


@attr.s(auto_attribs=True)
class UtilitarianConfigurationObject:
    parameter_type: int
    data: Dict
    series_name: Optional[str] = attr.ib(default=None)


@attr.s(auto_attribs=True)
class UtilitarianConfiguration:
    objects: List[UtilitarianConfigurationObject]


class UtilitarianConfigurationObjectSchema(marshmallow.Schema):

    parameter_type = marshmallow.fields.Integer(required=True)
    data = marshmallow.fields.Dict(required=True)
    series_name = marshmallow.fields.String(required=True, allow_none=True)


class UtilitarianConfigurationSchema(marshmallow.Schema):
    objects = marshmallow.fields.List(
        marshmallow.fields.Nested(UtilitarianConfigurationObjectSchema)
    )



from typing import ClassVar, Iterable, Optional

import attr
import marshmallow
from marshmallow import post_load

from elgas.parameters.enumerations import ParameterObjectType
from elgas.utils import pop_many, pretty_text


@attr.s(auto_attribs=True)
class TimeWindow:
    object_type: ClassVar[ParameterObjectType] = ParameterObjectType.TIME_WINDOW

    number: int
    id: int
    bit_order_in_actual_values: int
    bit_order_in_data_archive_record: int
    bit_order_in_binary_archive_record: int

    bit_control: int
    in_binary_archive: bool
    in_data_archive: bool
    name: str
    rows_in_window: int
    rows: Iterable[str]
    action_during_change: Optional[int]
    text_log_0: Optional[str]
    text_log_1: Optional[str]

    @classmethod
    def from_bytes(cls, in_data: bytes):
        data = bytearray(in_data)
        number = int.from_bytes(pop_many(data, 2), "little")
        id = int.from_bytes(pop_many(data, 2), "little")
        bit_order_in_actual_values = int.from_bytes(pop_many(data, 2), "little")
        bit_order_in_data_archive_record = int.from_bytes(pop_many(data, 2), "little")
        bit_order_in_binary_archive_record = int.from_bytes(pop_many(data, 2), "little")

        bit_control = data.pop(0)
        in_binary_archive = bool(bit_control & 0b00000001)
        in_data_archive = bool(bit_control & 0b00000010)
        name = pretty_text(pop_many(data, 23))
        rows_in_window = data.pop(0)
        rows = list()
        for _ in range(0, rows_in_window):
            rows.append(pop_many(data, 10).hex())
        if data:
            action_during_change = data.pop(0)
            text_log_0 = pretty_text(pop_many(data, 13))
            text_log_1 = pretty_text(pop_many(data, 13))
        else:
            action_during_change = None
            text_log_0 = None
            text_log_1 = None

        assert not data

        return cls(
            number=number,
            id=id,
            bit_order_in_actual_values=bit_order_in_actual_values,
            bit_order_in_data_archive_record=bit_order_in_data_archive_record,
            bit_order_in_binary_archive_record=bit_order_in_binary_archive_record,
            bit_control=bit_control,
            in_binary_archive=in_binary_archive,
            in_data_archive=in_data_archive,
            name=name,
            rows_in_window=rows_in_window,
            rows=rows,
            action_during_change=action_during_change,
            text_log_0=text_log_0,
            text_log_1=text_log_1,
        )


class TimeWindowSchema(marshmallow.Schema):

    number = marshmallow.fields.Integer(required=True)
    id = marshmallow.fields.Integer(required=True)
    bit_order_in_actual_values = marshmallow.fields.Integer(required=True)
    bit_order_in_data_archive_record = marshmallow.fields.Integer(required=True)
    bit_order_in_binary_archive_record = marshmallow.fields.Integer(required=True)

    bit_control = marshmallow.fields.Integer(required=True)
    in_binary_archive = marshmallow.fields.Boolean(required=True)
    in_data_archive = marshmallow.fields.Boolean(required=True)
    name = marshmallow.fields.String(required=True)
    rows_in_window = marshmallow.fields.Integer(required=True)
    rows = marshmallow.fields.List(marshmallow.fields.String, required=True)
    action_during_change = marshmallow.fields.Integer(required=True, allow_none=True)
    text_log_0 = marshmallow.fields.String(required=True, allow_none=True)
    text_log_1 = marshmallow.fields.String(required=True, allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return TimeWindow(**data)



import struct

import attr
import marshmallow
from marshmallow import post_load

from elgas.utils import pop_many


@attr.s(auto_attribs=True)
class GasComposition:
    co2: float
    n2: float
    combustion_heat: float
    relative_density: float
    h2: float
    h2s: float
    he: float
    h2o: float
    o2: float
    ar: float
    co: float
    c1h4: float
    c2h6: float
    c3h8: float
    ic4h10: float
    nc4h10: float
    ic5h12: float
    nc5h12: float
    c6h14: float
    c7h16: float
    c8h18: float
    c9h20: float
    c10h22: float

    @classmethod
    def from_bytes(cls, in_data: bytes):
        data = bytearray(in_data)

        return cls(
            co2=struct.unpack("<f", pop_many(data, 4))[0],
            n2=struct.unpack("<f", pop_many(data, 4))[0],
            combustion_heat=struct.unpack("<f", pop_many(data, 4))[0],
            relative_density=struct.unpack("<f", pop_many(data, 4))[0],
            h2=struct.unpack("<f", pop_many(data, 4))[0],
            h2s=struct.unpack("<f", pop_many(data, 4))[0],
            he=struct.unpack("<f", pop_many(data, 4))[0],
            h2o=struct.unpack("<f", pop_many(data, 4))[0],
            o2=struct.unpack("<f", pop_many(data, 4))[0],
            ar=struct.unpack("<f", pop_many(data, 4))[0],
            co=struct.unpack("<f", pop_many(data, 4))[0],
            c1h4=struct.unpack("<f", pop_many(data, 4))[0],
            c2h6=struct.unpack("<f", pop_many(data, 4))[0],
            c3h8=struct.unpack("<f", pop_many(data, 4))[0],
            ic4h10=struct.unpack("<f", pop_many(data, 4))[0],
            nc4h10=struct.unpack("<f", pop_many(data, 4))[0],
            ic5h12=struct.unpack("<f", pop_many(data, 4))[0],
            nc5h12=struct.unpack("<f", pop_many(data, 4))[0],
            c6h14=struct.unpack("<f", pop_many(data, 4))[0],
            c7h16=struct.unpack("<f", pop_many(data, 4))[0],
            c8h18=struct.unpack("<f", pop_many(data, 4))[0],
            c9h20=struct.unpack("<f", pop_many(data, 4))[0],
            c10h22=struct.unpack("<f", pop_many(data, 4))[0],
        )


class GasCompositionSchema(marshmallow.Schema):

    co2 = marshmallow.fields.Float(required=True, as_string=True)
    n2 = marshmallow.fields.Float(required=True, as_string=True)
    combustion_heat = marshmallow.fields.Float(required=True, as_string=True)
    relative_density = marshmallow.fields.Float(required=True, as_string=True)
    h2 = marshmallow.fields.Float(required=True, as_string=True)
    h2s = marshmallow.fields.Float(required=True, as_string=True)
    he = marshmallow.fields.Float(required=True, as_string=True)
    h2o = marshmallow.fields.Float(required=True, as_string=True)
    o2 = marshmallow.fields.Float(required=True, as_string=True)
    ar = marshmallow.fields.Float(required=True, as_string=True)
    co = marshmallow.fields.Float(required=True, as_string=True)
    c1h4 = marshmallow.fields.Float(required=True, as_string=True)
    c2h6 = marshmallow.fields.Float(required=True, as_string=True)
    c3h8 = marshmallow.fields.Float(required=True, as_string=True)
    ic4h10 = marshmallow.fields.Float(required=True, as_string=True)
    nc4h10 = marshmallow.fields.Float(required=True, as_string=True)
    ic5h12 = marshmallow.fields.Float(required=True, as_string=True)
    nc5h12 = marshmallow.fields.Float(required=True, as_string=True)
    c6h14 = marshmallow.fields.Float(required=True, as_string=True)
    c7h16 = marshmallow.fields.Float(required=True, as_string=True)
    c8h18 = marshmallow.fields.Float(required=True, as_string=True)
    c9h20 = marshmallow.fields.Float(required=True, as_string=True)
    c10h22 = marshmallow.fields.Float(required=True, as_string=True)

    @post_load
    def make_object(self, data, **kwargs):
        return GasComposition(**data)
